"""
This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

from typing import Dict, Union

import regex
from PyInquirer import prompt
from prettytable import PrettyTable, NONE
from prompt_toolkit.validation import Validator, ValidationError

from vaccination.core.task.base import BaseTask
from vaccination.service.api.booking import BookingAPIService


class _NumberValidator(Validator):
    def __init__(self):
        self.pattern = r""
        self.message = ""

    def validate(self, document):
        if not regex.match(self.pattern, document.text):
            raise ValidationError(
                message=self.message, cursor_position=len(document.text)
            )


class _PersonalNumberValidator(_NumberValidator):
    def __init__(self):
        super().__init__()
        self.pattern = r"^\d{11}$"
        self.message = "პირადი ნომერი არასწორია"


class _BookingNumberValidator(_NumberValidator):
    def __init__(self):
        super().__init__()
        self.pattern = r"^\d{6}$"
        self.message = "ჯავშნის ნომერი არასწორია"


class VaccinationCheckTask(BaseTask):
    """
    Vaccination Check CLI Task.
    """

    def __init__(self):
        self.api_service = BookingAPIService()
        self.steps = [
            [self._ask_personal_number, {}],
            [self._ask_booking_number, {}],
            [self._print_result, {}],
        ]

    def _ask_number(self, message: str, validator) -> str:
        answers = prompt(
            {
                "type": "input",
                "name": "number",
                "message": message,
                "validate": validator,
            },
            style=self.style,
        )

        number = answers.get("number")
        if not number:
            raise InterruptedError

        return number

    def _ask_personal_number(self) -> Dict[str, str]:
        return {
            "personal_number": self._ask_number(
                "აკრიფეთ პირადი ნომერი", _PersonalNumberValidator
            )
        }

    def _ask_booking_number(self, personal_number: str) -> Dict[str, str]:
        return {
            "personal_number": personal_number,
            "booking_number": self._ask_number(
                "აკრიფეთ ჯავშნის ნომერი", _BookingNumberValidator
            ),
        }

    def _print_result(
        self, personal_number: str = None, booking_number: str = None
    ) -> Union[Dict, None]:
        result = self.api_service.search_booking(personal_number, booking_number)

        if "value" in result and result["value"]:
            result = result["value"]
            table = PrettyTable(
                ["", " "],
                header=False,
                hrules=NONE,
                vrules=NONE,
                align="l",
            )

            table.add_rows(
                [
                    ["პიროვნება", f'- \t {result["firstName"]} {result["lastName"]}'],
                    ["დაბადების წელი", f'- \t {result["birthYear"]}'],
                    ["პირადი ნომერი", f'- \t {result["personalID"]}'],
                    ["ტელეფონის ნომერი", f'- \t {result["phone"]}'],
                    ["", ""],
                    ["სერვისი", f'- \t {result["testName"]}'],
                    ["დაწესებულება", f'- \t {result["branchName"]}'],
                    ["ოთახი", f'- \t {result["roomNumber"]}'],
                    ["თარიღი/დრო", f'- \t {result["scheduleDateName"]}'],
                ]
            )

            print(f"\n{table}\n")
        else:
            print(f"\033[91m{result['message']}\033[0m")

        return {}



"""
This module contains the API service for working with stopcov-api.lotto.ge's API.

This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

from vaccination.service.api.base import BaseAPIService


class LottoAPIService(BaseAPIService):
    """
    Service for working with the stopcov-api.lotto.ge's API.
    """

    url_template = "https://stopcov-api.lotto.ge$path"

    def check_winning(self, personal_number: str) -> bool:
        """
        Check Lotto results for given Personal ID.

        :param str personal_number: Personal ID to check results for.
        :return: Winning status.
        """

        return self._get(url={"path": f"/Public/Winnings/{personal_number}"})



"""
Abstract API service.

This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""


from string import Template

import requests
from requests.models import Response


def retry_request(times):
    """
    Request Retry Decorator.

    :param times: The number of times to repeat the wrapped function/method.
    :return:
    """

    def decorator(function):
        def wrapper(*args, **kwargs):
            attempt = 0
            while attempt < times:
                response = function(*args, **kwargs)
                if response.status_code == 404:
                    attempt += 1
                    continue

                return response

            return function(*args, **kwargs)

        return wrapper

    return decorator


class BaseAPIService:
    """
    Base service for working with the APIs.
    """

    url_template = None

    @retry_request(times=20)
    def _make_request(self, method: str, **kwargs) -> Response:
        url = Template(self.url_template).substitute(**kwargs.get("url", {}))
        del kwargs["url"]
        return requests.request(method, url, **kwargs)

    def _get(self, **kwargs):
        return self._make_request("get", **kwargs).json()

    def _post(self, **kwargs):
        return self._make_request("post", **kwargs).json()



"""
This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

from typing import List, Dict

from PyInquirer import style_from_dict, Token, Separator, prompt


class BaseTask:
    """
    Base CLI Task.
    """

    style = style_from_dict(
        {
            Token.Separator: "#6C6C6C",
            Token.QuestionMark: "#FF9D00 bold",
            Token.Selected: "#5F819D",
            Token.Pointer: "#FF9D00 bold",
            Token.Instruction: "",
            Token.Answer: "#5F819D bold",
            Token.Question: "",
        }
    )
    back_choice = "<< უკან დაბრუნება"
    steps = []

    def _dict_to_choices(
        self, raw: Dict[str, any], navigation: bool = True
    ) -> List[str]:
        choices = list(raw.keys())
        return (
            choices + [Separator("-" * 18), self.back_choice] if navigation else choices
        )

    def _ask_to_retry(self, message: str, default: bool = False) -> bool:
        answers = prompt(
            {
                "type": "confirm",
                "message": message,
                "name": "retry",
                "default": default,
            },
            style=self.style,
        )

        return answers.get("retry")

    def run(self) -> int:
        """
        Run task.

        :return: Exit code.
        """

        i = 0
        while i < len(self.steps):
            input_data = self.steps[i - 1][1] if i != 0 else {}

            try:
                output_data = self.steps[i][0](**input_data)
            except InterruptedError:
                return 0

            if output_data is None:
                i = i - 1
                continue

            self.steps[i][1] = output_data
            i = i + 1

        return 0



"""
This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

from typing import Dict, Union

import regex
from PyInquirer import prompt
from prompt_toolkit.validation import Validator, ValidationError

from vaccination.core.task.base import BaseTask
from vaccination.service.api.lotto import LottoAPIService


class _PersonalNumberValidator(Validator):
    def validate(self, document):
        if not regex.match(r"^\d{11}$", document.text):
            raise ValidationError(
                message="პირადი ნომერი არასწორია", cursor_position=len(document.text)
            )


class LottoTask(BaseTask):
    """
    Lotto CLI Task.
    """

    def __init__(self):
        self.api_service = LottoAPIService()
        self.steps = [
            [self._ask_personal_number, {}],
            [self._print_result, {}],
        ]

    def _ask_personal_number(self) -> Dict[str, str]:
        answers = prompt(
            {
                "type": "input",
                "name": "personal_number",
                "message": "აკრიფეთ პირადი ნომერი",
                "validate": _PersonalNumberValidator,
            },
            style=self.style,
        )

        personal_number = answers.get("personal_number")
        if not personal_number:
            raise InterruptedError

        return {"personal_number": personal_number}

    def _print_result(self, personal_number: str) -> Union[Dict, None]:
        winning = self.api_service.check_winning(personal_number)

        if winning:
            print("\033[92mგილოცავთ, თქვენ გაიმარჯვეთ!\033[0m")
        else:
            print("\033[91mსამწუხაროდ, თქვენ ვერ გაიმარჯვეთ!\033[0m")

        return None if self._ask_to_retry("გსურთ სხვა მონაცემის შემოწმება?") else {}



"""
Main application module.

This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

from vaccination.core.task.main import MainTask


class App:
    """
    Application class.
    """

    @staticmethod
    def run() -> int:
        """
        Run application.

        :return: Exit code.
        """

        return MainTask().run()



"""
This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

from vaccination import __version__, __copyright__, __email__
from vaccination.core.app import App


def __print_banner() -> None:
    print("__     __             _             _   _             ")
    print(r"\ \   / /_ _  ___ ___(_)_ __   __ _| |_(_) ___  _ __  ")
    print(r" \ \ / / _` |/ __/ __| | '_ \ / _` | __| |/ _ \| '_ \ ")
    print(r"  \ V / (_| | (_| (__| | | | | (_| | |_| | (_) | | | |")
    print(r"   \_/ \__,_|\___\___|_|_| |_|\__,_|\__|_|\___/|_| |_|")
    print(f"\n     v{__version__} {__copyright__} ({__email__})\n")


def main() -> int:
    """
    Main function.
    :return: Exit code.
    """

    __print_banner()
    return App.run()


if __name__ == "__main__":
    import sys

    sys.exit(main())



"""
This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

import os
import re

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def get_requirements(filename: str) -> list[str]:
    """
    Get requirements file content by filename.

    :param filename: Name of requirements file.
    :return: Content of requirements file.
    """

    return open("requirements/" + filename).read().splitlines()


def get_package_version() -> str:
    """
    Read the version of Vaccination module without importing it.

    :return: The version.
    """

    version = re.compile(r"__version__\s*=\s*\"(.*?)\"")
    base = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(base, "vaccination/__init__.py")) as file:
        for line in file:
            match = version.match(line.strip())
            if not match:
                continue
            return match.groups()[0]


setup(
    name="vaccination",
    version=get_package_version(),
    author="Temuri Takalandze",
    author_email="me@abgeo.dev",
    description="Get COVID-19 vaccination schedules from booking.moh.gov.ge in the CLI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ABGEO/vaccination-py",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=get_requirements("default.txt"),
    entry_points={
        "console_scripts": [
            "vaccination = vaccination.__main__:main",
        ]
    },
)



"""
This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

from typing import Dict

from PyInquirer import prompt

from vaccination.core.task.base import BaseTask
from vaccination.core.task.lotto import LottoTask
from vaccination.core.task.vaccination import VaccinationTask
from vaccination.core.task.vaccination_check import VaccinationCheckTask


class MainTask(BaseTask):
    """
    Main CLI Task.
    """

    def __init__(self):
        self.steps = [
            [self._select_task, {}],
            [self._run_task, {}],
        ]

    def _select_task(self) -> Dict[str, callable]:
        tasks = {
            "ვაქცინაცია": VaccinationTask,
            "ჯავშნის შემოწმება": VaccinationCheckTask,
            "ლოტო": LottoTask,
        }
        answers = prompt(
            {
                "type": "list",
                "name": "task",
                "message": "აირჩიეთ სერვისი",
                "choices": self._dict_to_choices(tasks, False),
            },
            style=self.style,
        )

        task = answers.get("task")
        if not task:
            raise InterruptedError

        return {"task": tasks[task]}

    @staticmethod
    def _run_task(task: callable) -> Dict[str, str]:
        return task().run()



"""
This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

import copy
import datetime
import os
from datetime import date
from typing import List, Dict, Union

from PyInquirer import prompt
from prettytable import PrettyTable, ALL

from vaccination.core.task.base import BaseTask
from vaccination.service.api.booking import BookingAPIService


class VaccinationTask(BaseTask):
    """
    Vaccination CLI Task.
    """

    def __init__(self):
        self.api_service = BookingAPIService()
        self.steps = [
            [self._select_service, {}],
            [self._select_region, {}],
            [self._select_municipality, {}],
            [self._select_branch, {}],
            [self._select_room, {}],
            [self._print_result, {}],
        ]

    def _select_service(self) -> Dict[str, str]:
        quantities = self.api_service.get_available_quantities()
        services = {}
        for app in ["abc", "def"]:
            for service in self.api_service.get_service_types(app):
                key = service["name"]
                key = key[key.find("(") + 1 : key.find(")")]
                key += f" ({quantities[key.lower()]:,})"
                services[key] = service["id"]

        answers = prompt(
            {
                "type": "list",
                "name": "service",
                "message": "აირჩიეთ ვაქცინა",
                "choices": self._dict_to_choices(services, False),
            },
            style=self.style,
        )

        service = answers.get("service")
        if not service:
            raise InterruptedError

        regions = {}
        for region in self.api_service.get_regions(services[service]):
            regions[region["geoName"]] = region["id"]

        return {"service": services[service], "regions": regions}

    def _select_region(
        self, service: str, regions: Dict[str, str]
    ) -> Union[Dict[str, Union[str, Dict[str, str]]], None]:
        answers = prompt(
            {
                "type": "list",
                "name": "region",
                "message": "სერვისის ჩატარების რეგიონი",
                "choices": self._dict_to_choices(regions),
            },
            style=self.style,
        )

        region = answers.get("region")
        if not region:
            raise InterruptedError

        if region == self.back_choice:
            return None

        municipalities = {}
        for municipality in self.api_service.get_municipalities(
            regions[region], service
        ):
            municipalities[municipality["geoName"]] = municipality["id"]

        return {
            "service": service,
            "region": regions[region],
            "municipalities": municipalities,
        }

    def _select_municipality(
        self, service: str, region: str, municipalities: Dict[str, str]
    ) -> Union[Dict[str, Union[str, Dict[str, str]]], None]:
        answers = prompt(
            {
                "type": "list",
                "name": "municipality",
                "message": "სერვისის ჩატარების რაიონი",
                "choices": self._dict_to_choices(municipalities),
            },
            style=self.style,
        )

        municipality = answers.get("municipality")
        if not municipality:
            raise InterruptedError

        if municipality == self.back_choice:
            return None

        branches = {}
        for branch in self.api_service.get_municipality_branches(
            service, municipalities[municipality]
        ):
            branches[branch["name"]] = branch["id"]

        return {
            "region": region,
            "service": service,
            "branches": branches,
        }

    def _select_branch(
        self, region: str, service: str, branches: Dict[str, str]
    ) -> Union[Dict[str, Dict[str, List]], None]:
        answers = prompt(
            {
                "type": "list",
                "name": "branch",
                "message": "სერვისის მიმწოდებელი დაწესებულება",
                "choices": self._dict_to_choices(branches),
            },
            style=self.style,
        )

        branch = answers.get("branch")
        if not branch:
            raise InterruptedError

        if branch == self.back_choice:
            return None

        start_date = date.today()
        end_date = start_date + datetime.timedelta(days=7)

        rooms = {}
        for room in self.api_service.get_slots(
            branches[branch], region, service, start_date, end_date
        ):
            rooms[room["name"]] = room["schedules"]

        return {"rooms": rooms}

    def _select_room(self, rooms: Dict[str, List]) -> Union[Dict[str, List], None]:
        answers = prompt(
            {
                "type": "list",
                "name": "room",
                "message": "აირჩიეთ კაბინეტი",
                "choices": self._dict_to_choices(rooms),
            },
            style=self.style,
        )

        room = answers.get("room")
        if not room:
            raise InterruptedError

        if room == self.back_choice:
            return None

        dates = copy.deepcopy(rooms[room][0]["dates"])
        for i, item in enumerate(dates):
            dates[i]["slots"] = list(map(lambda x: x["value"], item["slots"]))

        return {"dates": dates}

    def _print_result(self, dates: List) -> Union[Dict, None]:
        _, columns = os.popen("stty size", "r").read().split()
        header = ["თარიღი", "დღე", "თავისუფალი დროები"]
        table = PrettyTable(
            header,
            hrules=ALL,
            align="l",
            min_width=60,
            max_table_width=int(columns) - 10,
        )
        table.add_rows(
            map(lambda x: [x["dateName"], x["weekName"], ", ".join(x["slots"])], dates)
        )

        print(f"\n{table}\n")

        return None if self._ask_to_retry("გსურთ სხვა კაბინეტის ნახვა?") else {}



"""
This file is part of the vaccination.py.

(c) 2021 Temuri Takalandze <me@abgeo.dev>

For the full copyright and license information, please view the LICENSE
file that was distributed with this source code.
"""

__project__ = "vaccination"
__author__ = "Temuri Takalandze"
__copyright__ = "(c) 2021 Temuri Takalandze"
__email__ = "me@abgeo.dev"
__credits__ = ["Temuri Takalandze"]
__license__ = "MIT"
__version__ = "1.2.0"
__status__ = "Stable"
